#
# Copyright (c) 2003 Art Haas
#
# This file is part of PythonCAD.
#
# PythonCAD is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PythonCAD is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PythonCAD; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# class stuff for ellipses
#
# Ellipse info:
# http://mathworld.wolfram.com/Ellipse.html
# http://astronomy.swin.edu.au/~pbourke/geometry/ellipsecirc/
#

import math

import baseobject
import tolerance
import point
import graphicobject
import style
import linetype
import color
import util

class _Ellipse(object):
    """A base class for Ellipses

An ellipse has the following attributes:

center: A _point object
major_axis:
minor_axis:
angle: A float from -90.0 to 90.0 degrees
    """
    def __init__(self, center, major_axis, minor_axis, angle):
        _cp = center
        if not isinstance(_cp, point._Point):
            _cp = point.Point(center)
        _major = major_axis
        if not isinstance(_major, float):
            _major = float(major_axis)
        if not _major > 0.0:
            raise ValueError, "Invalid major axis value: %g" % _major
        _minor = minor_axis
        if not isinstance(_minor, float):
            _minor = float(minor_axis)
        if not _minor > 0.0:
            raise ValueError, "Invalid minor axis value: %g" % _minor
        if _minor > _major:
            raise ValueError, "Minor axis must be less than major axis"
        _angle = angle
        if not isinstance(_angle, float):
            _angle = float(angle)
        _angle = util.make_angle(_angle)
        self.__center = _cp
        self.__major = _major
        self.__minor = _minor
        self.__angle = _angle

    def __eq__(self, obj):
        """Compare one ellipse to another for equality.
        """
        _val = False
        if isinstance(obj, _Ellipse):
            if (self.__center == obj.getCenter() and
                abs(self.__major - obj.getMajorAxis()) < 1e-10 and
                abs(self.__minor - obj.getMinorAxis()) < 1e-10 and
                abs(self.__angle - obj.getAngle()) < 1e-10):
                _val = True
        return _val

    def __ne__(self, obj):
        """Compare one ellipse to another for equality.
        """
        _val = True
        if isinstance(obj, _Ellipse):
            if (self.__center == obj.getCenter() and
                abs(self.__major - obj.getMajorAxis()) < 1e-10 and
                abs(self.__minor - obj.getMinorAxis()) < 1e-10 and
                abs(self.__angle - obj.getAngle()) < 1e-10):
                _val = False
        return _val


    def getCenter(self):
        """Return the center _Point of the _Ellipse.

getCenter()
        """
        return self.__center

    def setCenter(self, cp):
        """Set the center _Point of the _Ellipse.

setCenter(cp)

The argument must be a _Point or a tuple containing
two float values.
        """
        _cp = cp
        if not isinstance(_cp, point._Point):
            _cp = point._Point(cp)
        self.__center = _cp

    center = property(getCenter, setCenter, None, "_Ellipse center")

    def getMajorAxis(self):
        """Return the major axis value of the _Ellipse.

getMajorAxis()

This method returns a float.
        """
        return self.__major

    def setMajorAxis(self, val):
        """Set the major axis of the _Ellipse.

setMajorAxis(val)

Argument 'val' must be a float value greater than 0.
        """
        _major = val
        if not isinstance(_major, float):
            _major = float(val)
        if not _major > 0.0:
            raise ValueError, "Invalid major axis value: %g" % _major
        if _major < self.__minor:
            raise ValueError, "Major axis must be greater than minor axis."
        self.__major = _major

    major_axis = property(getMajorAxis, setMajorAxis, None,
                          "_Ellipse major axis")

    def getMinorAxis(self):
        """Return the minor axis value of the _Ellipse.

getMinorAxis()

This method returns a float.
        """
        return self.__minor

    def setMinorAxis(self, val):
        """Set the minor axis of the _Ellipse.

setMinorAxis(val)

Argument 'val' must be a float value greater than 0.
        """
        _minor = val
        if not isinstance(_minor, float):
            _minor = float(val)
        if not _minor > 0.0:
            raise ValueError, "Invalid minor axis value: %g" % _minor
        if _minor > self.__major:
            raise ValueError, "Minor axis must be less than major axis"
        self.__minor = _minor

    minor_axis = property(getMinorAxis, setMinorAxis, None,
                          "_Ellipse minor axis")

    def getAngle(self):
        """Return the _Ellipse major axis angle.

getAngle()

This method returns a float value.
        """
        return self.__angle

    def setAngle(self, angle):
        """Set the _Ellipse major axis angle.

setAngle(angle)

Argument 'angle' should be a float. The value will be
adjusted so the angle will be defined from 90.0 to -90.0.
        """
        _angle = angle
        if not isinstance(_angle, float):
            _angle = float(angle)
        _angle = util.make_angle(_angle)

    angle = property(getAngle, setAngle, None, "_Ellipse major axis angle")

    def move(self, dx, dy):
        """Move an _Ellipse.

move(dx, dy)

Arguments 'dx' and 'dy' should be floats.
        """
        _dx = dx
        if not isinstance(_dx, float):
            _dx = float(dx)
        _dy = dy
        if not isinstance(_dy, float):
            _dy = float(dy)
        self.__center.move(_dx, _dy)

    def rotate(self, angle):
        """Rotate an _Ellipse

rotate(angle)

Argument 'angle' should be a float.
        """
        _angle = angle
        if not isinstance(_angle, float):
            _angle = float(_angle)
        _cur = self.__angle
        self.__angle = util.make_angle(_angle + _cur)

    def eccentricity(self):
        """Return the eccecntricity of the _Ellipse.

eccentricity()

This method returns a float value.
        """
        _major = self.__major
        _minor = self.__minor
        if abs(_major - _minor) < 1e-10: # circular
            _e = 0.0
        else:
            _e = math.sqrt(1.0 - ((_minor * _minor)/(_major * _major)))
        return _e

    def area(self):
        """Return the area of the _Ellipse.

area()

This method returns a float value.
        """
        return math.pi * self.__major * self.__minor

    def circumference(self):
        """Return the circumference of an ellipse.

circumference()

This method returns a float.

The algorithm below is taken from
http://astronomy.swin.edu.au/~pbourke/geometry/ellipsecirc/
Ramanujan, Second Approximation
        """
        _a = self.__major
        _b = self.__minor
        _h = math.pow((_a - _b), 2)/math.pow((_a + _b), 2)
        _3h = 3.0 * _h
        return math.pi * (_a + _b) * (1.0 + _3h/(10.0 + math.sqrt(4.0 - _3h)))

    def mapCoords(self, x, y, tol=tolerance.TOL):
        """Return the nearest _Point on the _Ellipse to a coordinate pair.

mapCoords(x, y[, tol])

The function has two required arguments:

x: A Float value giving the "x" coordinate
y: A Float value giving the "y" coordinate

There is a single optional argument:

tol: A float value equal or greater than 0.0

This function is used to map a possibly near-by coordinate pair to
an actual _Point on the _Ellipse. If the distance between the actual
_Point and the coordinates used as an argument is less than the tolerance,
the actual _Point is returned. Otherwise, this function returns None.
        """
        _x = x
        if not isinstance(_x, float):
            _x = float(x)
        _y = y
        if not isinstance(_y, float):
            _y = float(y)
        _t = tolerance.toltest(tol)
        _cx, _cy = self.__center.getCoords()
        _dist = math.hypot((_x - _cx), (_y - _cy))
        _major = self.__major
        _minor = self.__minor
        _ep = None
        if abs(_major - _minor) < 1e-10: # circular
            _sep = _dist - _major
            if _sep < _t or abs(_sep - _t) < 1e-10:
                _angle = math.atan2((_y - _cy),(_x - _cx))
                _px = _major * math.cos(_angle)
                _py = _major * math.sin(_angle)
                _ep = point._Point((_cx + _px), (_cy + _py))
        else:
            if _dist < _major and _dist > _minor:
                _ecos = math.cos(self.__angle)
                _esin = math.sin(self.__angle)
        return _ep
#
# measure r from focus
#
# x = c + r*cos(theta)
# y = r*sin(theta)
#
# c = sqrt(a^2 - b^2)
#
# r = a*(1-e)/(1 + e*cos(theta))

    def clone(self):
        """Make a copy of an _Ellipse.

clone()

This method returns a new _Ellipse object
        """
        _cp = self.__center.clone()
        return _Ellipse(_cp, self.__major, self.__minor, self.__angle)

class Ellipse(_Ellipse, baseobject.Subpart, baseobject.ModObject,
               graphicobject.GraphicObject):
    """The storable and displayed representation of an _Ellipse

An Ellipse object is derived from an _Ellipse object, Subpart, ModObject,
and a GraphicObject, so it has the attributes and member functions
as in those objects.
    """

    #
    # The default style for the Ellipse class
    #

    __defstyle = style.Style(u'Default Ellipse Style',
                            linetype.Linetype(u'Solid', None),
                            color.Color(0xffffff),
                            1.0)
    # member functions

    def __init__(self, center, major, minor, angle,
                 st=None, lt=None, col=None, t=None):
        """Initialize an Ellipse object.

Ellipse(center, major, minor, angle[, st, lt, col, t])

There are four required arguments:

center: A Point object for the Ellipse's center.
major: A float giving the Ellipse's major axis.
minor: A float giving the Ellipse's minor axis.
angle: A float giving the angle at which the major axis is inclided.

Optional arguments:

st: A Style object
lt: A Linetype object that overrides the Style linetype
col: A Color object that overrides the Style color
t: A float that overrides the Style thickness
        """
        _center = center
        if not isinstance(_center, point.Point):
            _center = point.Point(center)
        _st = st
        if _st is None:
            _st = Ellipse.__defstyle
        _Ellipse.__init__(self, _center, major, minor, angle)
        baseobject.ModObject.__init__(self)
        baseobject.Subpart.__init__(self)
        graphicobject.GraphicObject.__init__(self, _st, lt, col, t)
        self.reset()
        _center.storeUser(self)

    def __str__(self):
        _cp = self.getCenter()
        return "Ellipse: Center: %s; Major: %g; Minor: %g; Angle: %g" % (str(_cp), self.__major, self.__minor, self.__angle)

    def isModified(self):
        """Indicate if the Ellipse has been changed.

isModified()

This method extends the ModObject::isModified() method.
        """
        if self.getCenter().isModified():
            self.modified()
        return baseobject.ModObject.isModified(self)

    def setCenter(self, c):
        """Set the center Point of the Ellipse

setCenter(c)

This method extends the _Ellipse::setCenter() method and sets
the modified flag indicating the Ellipse has been altered.
        """
        if not isinstance(c, point.Point):
            raise TypeError, "Invalid point: " + `c`
        _cp = self.getCenter()
        _Ellipse.setCenter(self, c)
        self.modified()
        _cp.freeUser(self)
        c.storeUser(self)

    center = property(_Ellipse.getCenter, setCenter, None,
                      "Ellipse center point.")

    def setMajorAxis(self, val):
        """Set the major axis of the Ellipse.

setMajorAxis(val)

This method extends the _Ellipse::setMajorAxis() method and sets
the modified flag indicating the Ellipse has been altered.
        """
        _Ellipse.setMajorAxis(self, val)
        self.modified()

    major_axis = property(_Ellipse.getMajorAxis, setMajorAxis, None,
                          "Ellipse major axis")

    def setMinorAxis(self, val):
        """Set the minor axis of the Ellipse.

setMinorAxis(val)

This method extends the _Ellipse::setMinorAxis() method and sets
the modified flag indicating the Ellipse has been altered.
        """
        _Ellipse.setMinorAxis(self, val)
        self.modified()

    minor_axis = property(_Ellipse.getMinorAxis, setMinorAxis, None,
                          "Ellipse minor axis")

    def setAngle(self, angle):
        """Set the Ellipse major axis angle.

setAngle(angle)

This method extends the _Ellipse::setAngle() method and sets
the modified flag indicating the Ellipse has been altered.
        """
        _Ellipse.setAngle(self, angle)
        self.modified()

    angle = property(_Ellipse.getAngle, setAngle, None,
                     "Ellipse major axis angle")

    def move(self, dx, dy):
        """Move an Ellipse.

move(dx, dy)

This method extends the _Ellipse::move() method and sets
the modified flag indicating the Ellipse has been altered.
        """
        _Ellipse.move(self, dx, dy)
        self.modified()

    def rotate(self, angle):
        """Rotate an Ellipse.

rotate(angle)

This method extends the _Ellipse::rotate() method and sets
the modified flag indicating the Ellipse has been altered.
        """
        _Ellipse.rotate(self, angle)
        self.modified()
